package Rione_2021.module.complex.taskManager;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.communication.standard.bundle.information.MessageRoad;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.algorithm.Clustering;
import adf.component.module.complex.RoadDetector;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

/**
 * @author Horie(16th)
 * @since 2020.10.- summary : 動的なtarget選択のために, Comparatorと優先度管理クラスを実装 優先度と,
 *        優先度の変化しやすさ（＝不安定度）を各対象ごとに求め管理する
 * 
 */

public class RIODynamicTaskManager {

    //// attention : オーバーフロー対策必要？
    //private int insertOrder = 0; // 挿入順

    /// 情報
    private ScenarioInfo scenarioInfo;
    private WorldInfo    worldInfo;
    private AgentInfo    agentInfo;
    private PathPlanning pathPlanning;
    private Clustering   clustering;

    /// target候補管理
    private Set<RIODynamicTaskEvaluator> evaluators = new HashSet<RIODynamicTaskEvaluator>();
    private Set<RIODynamicTaskKey> prioritySet = new HashSet<RIODynamicTaskKey>();
    private PriorityQueue<RIODynamicTaskKey> unstablenessQueue = new PriorityQueue<RIODynamicTaskKey>(
            new UnstablenessChecker());
    
    /// ---- 内部クラス ----
    /// summary : 不安定度管理のための優先度付きキュー管理クラス
    private class UnstablenessChecker implements Comparator<RIODynamicTaskKey> {
        public int compare(RIODynamicTaskKey key1, RIODynamicTaskKey key2) {
            // 不安定度を比較
            if (key1.unstableness < key2.unstableness){
                return 1;
            }
            if (key1.unstableness > key2.unstableness){
                return -1;
            }
            return 0;
        }
    }

    /// summary : target決定のための優先度付きキュー管理クラス
    private class PriorityChecker implements Comparator<RIODynamicTaskKey> {
        public int compare(RIODynamicTaskKey key1, RIODynamicTaskKey key2) {
            // 評価値を比較
            if (key1.priority < key2.priority){
                return 1;
            }
            if (key1.priority > key2.priority){
                return -1;
            }
            return 0;
        }
    }

    /// ---- メンバメソッド ----
    /// この順番に実行しないと実行時エラーを招きます（要検討）
    /// summary : ビルダーメソッド（Info系）
    public RIODynamicTaskManager setInfo(ScenarioInfo scenarioInfo, 
                                         WorldInfo worldInfo,
                                         AgentInfo agentInfo       ){
        // 情報の取得
        this.scenarioInfo = scenarioInfo;
        this.worldInfo = worldInfo;
        this.agentInfo = agentInfo;
        return this;
    }

    /// summary ビルダーメソッド（アルゴリズム系）
    public RIODynamicTaskManager setAlgorithm(PathPlanning pathPlanning, Clustering clustering){
        this.pathPlanning = pathPlanning;
        this.clustering   = clustering;
        return this;
    }
    
    /// summary : ビルダーメソッド（評価対象）
    public RIODynamicTaskManager setTargetKind(StandardEntityURN type){
        // prioritySetのセットアップ
        Collection<EntityID> allTargets = this.worldInfo.getEntityIDsOfType(type);
        for(EntityID id : allTargets){
            this.insert(id);
        }
        return this;
    }

    /// summary : 評価基準の追加
    public RIODynamicTaskManager setEvaluator(RIODynamicTaskEvaluator evaluator) {
        evaluator.setInfo(this.scenarioInfo, this.worldInfo, this.agentInfo)
                 .setAlgorithm(this.pathPlanning, this.clustering);
        this.evaluators.add(evaluator);
        return this;
    }

    /// summary : キーの挿入, target候補への評価値等付与
    private RIODynamicTaskManager insert(EntityID keyID) {
        // 時間経過による不安定度の低減（編集中）


        // 評価値と不安定度の計算
        double priority = 0;
        double unstableness = 0;
        for (RIODynamicTaskEvaluator taskEvaluator : evaluators) {
            taskEvaluator.setEvaluatee(keyID);
            priority += taskEvaluator.getPriority();
            unstableness += taskEvaluator.getUnstableness();
        }

        // 各コレクションへのタスクの登録
        RIODynamicTaskKey nowKey = new RIODynamicTaskKey(keyID, priority, unstableness);
        this.prioritySet.add(nowKey);
        this.unstablenessQueue.add(nowKey);
        return this;
    }

    /// summary : 不安定度が高いキーの優先度再評価
    public RIODynamicTaskManager update() {
        // 各Evaluatorのupdate
        for (RIODynamicTaskEvaluator taskEvaluator : evaluators) {
            taskEvaluator.update();
        }

        // コレクションからの取り出し・削除
        RIODynamicTaskKey updateeKey = this.unstablenessQueue.poll();
        prioritySet.remove(updateeKey);

        // 評価値の再付与
        this.insert(updateeKey.key);
        return this;
    }

    /// summary : target決定処理の実行
    //// 取り出し後の処理検討中
    public EntityID executation() {
        RIODynamicTaskKey ans = Collections.max(prioritySet, new PriorityChecker());
        return ans.key;
    }
}