package Rione_2021.module.complex.taskManager;

import adf.agent.Agent;
import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.communication.standard.bundle.information.MessageRoad;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.algorithm.Clustering;
import adf.component.module.complex.RoadDetector;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import Rione_2021.module.complex.taskManager.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

/**
 * @author Horie(16th)
 * @since 2020.10.-
 * summary : 各基準ごとに, 優先度と優先度の変わりやすさ（不安定度）を算出するクラス
 * 
 * 以下, 注意事項(by Horie)
 *
 * 1.  抽象クラスであるので, これを継承して評価値計算クラスを作る 
 * 2.  実際に使うときは, 具象クラスではなくRIODynamicTaskEvaluatorのインスタンスとして各インスタンスを定義する
 * 3.  bestPriority() < worstPriority()である必要はない.
 *     考えた評価基準の意味に沿って実装してくれればnormalizeがうまくやってくれる.
 * 4.  finalやabstractで制限をかけているものにはそれだけの理由があるので制限は外さないで！
 */

public abstract class RIOEvaluatorNormal implements RIODynamicTaskEvaluator {
    // 評価のための情報
    private boolean hasEvaluated;
    private RIODynamicTaskKey evaluateResult;

    protected ScenarioInfo scenarioInfo;
    protected WorldInfo    worldInfo;
    protected AgentInfo    agentInfo;
    protected PathPlanning pathPlanning;
    protected Clustering   clustering;

    // 評価対象（target候補）
    protected EntityID evaluateeID;

    /// --- 各種情報設定 ---
    /// summary : 優先度計算対象の設定, 継承禁止
    @Override
    public final RIODynamicTaskEvaluator setEvaluatee(EntityID evaluateeID){
        this.evaluateeID = evaluateeID;
        this.evaluateResult = this.calc();
        return this;
    }

    /// summary : エージェント側のプロパティの引き渡し, 継承禁止
    @Override
    public final RIODynamicTaskEvaluator setInfo(ScenarioInfo scenarioInfo,
                                                 WorldInfo    worldInfo, 
                                                 AgentInfo    agentInfo    ) {
        // 情報の取得
        this.scenarioInfo = scenarioInfo;
        this.worldInfo    = worldInfo;
        this.agentInfo    = agentInfo;
        this.init();
        return this;
    }

    @Override
    public final RIODynamicTaskEvaluator setAlgorithm(PathPlanning pathPlanning, Clustering clustering) {
        this.pathPlanning = pathPlanning;
        this.clustering   = clustering;
        return this;
    }

    /// summary 初期化後の処理
    protected RIODynamicTaskEvaluator init(){
        return this;
    }

    /// ---- 各種計算 ----
    /// summary : 各サイクルごとの事前処理
    @Override
    public RIODynamicTaskEvaluator update(){
        this.hasEvaluated = false;
        return this;
    }

    /// summary : 優先度を返す関数, 継承禁止
    @Override
    public final double getPriority(){
        return (this.evaluateResult.priority - bestPriority()) / (worstPriority() - bestPriority());
    }

    /// sumamry : 不安定度を返す関数, 継承禁止
    @Override
    public final double getUnstableness(){
        return (this.evaluateResult.unstableness - bestUnstableness()) / (worstUnstableness() - bestUnstableness());
    }

    /// summary : 優先度・不安定度を計算する関数, 継承必須
    protected abstract RIODynamicTaskKey calc();

    /// summary : 正規化の基準関数, 継承必須
    protected abstract double bestPriority();
    protected abstract double worstPriority();
    protected abstract double bestUnstableness();
    protected abstract double worstUnstableness();
}
